package moca.core.proxy.receivers;

import moca.core.proxy.ClientInfo;
import java.util.LinkedList;
/**
 * Classe base para todos os componentes processadores de pacotes. Um receiver possui
 * uma fila individual para a qual pacotes podem ser adicionadas. Os pacotes sero lidos
 * e processados na ordem em que foram inseridos (FIFO), de acordo com a capacidade de 
 * processamento do receiver
 * @author hana
*/

public abstract class Receiver {
	

	/**
	* Esta classe permite que o mtodo process possa retorna a mensagem o seu proximo
	* receiver de destino
	* @author hana
	*
	*/
	protected static class ReceiverResult {
		/**
		 * Mensagem retornada
		 */
		private ReceiverMessage msg;
		/**
		 * Proximo receiver que ir processa-la
		 */
		private Object dest;
		/**
		 * Construtor
		 * @param msg mensagem
		 * @param dest receiver destino
		 */
		public ReceiverResult(ReceiverMessage msg, Object dest) {
			this.dest = dest;
			this.msg = msg;
		}
		/**
		 * Recupera a mensagem
		 * @return mensagem
		 */
		public ReceiverMessage getMessage() {
			return this.msg;
		}
		/**
		 * Recupera o receiver de destino
		 * @return receiver destino
		 */
		public Object getDestination() {
			return this.dest;
		}
	}

	/**
	 * Referencia para o dispatcher
	 */
	protected Dispatcher disp;
	/**
	 * Seta o dispatcher
	 * @param disp dispatcher
	 */
	public void setDispatcher(Dispatcher disp) {
		this.disp = disp;
	}
	/**
	 * Fial de pacotes do receiver
	 */
	private LinkedList packageFiFo = new LinkedList();
	/**
	 * Adiciona um pacote a fila do receiver e notifica as threads
	 * bloqueadas
	 * @param pack pacote
	 */
	protected synchronized void addPackage(moca.core.proxy.receivers.Package pack) {
		this.packageFiFo.add(pack);
		this.notify();
	}
	/**
	 * Adiciona um array de pacotes para a fila do receiver
	 * e notifica as threads bloqueadas
	 * @param pack array de pacotes 
	 */
	protected synchronized void addPackages(Package[] pack) {
		for (int i = 0; i < pack.length; i++) {
			this.packageFiFo.add(pack[i]);
		}
		this.notifyAll();
	}
	
	/**
	 * Recupera um pacote da fila do receiver. Este mtodo
	 * bloqueia a thread caso no haja mensagens na fila
	 * @return
	 */
	protected synchronized moca.core.proxy.receivers.Package getNextPackage() {
		while (this.packageFiFo.size() == 0) {
			try {
				wait();
			} catch (InterruptedException e) {
				throw new RuntimeException(e);
			}
		}
		return (moca.core.proxy.receivers.Package) this.packageFiFo.removeFirst();		
	}

	
	/**
	 * Implementacao default para o receiver. Este mtodo
	 * processa cada uma das mensagens da fila
	 */
	public void process() {
		while (true) {
			moca.core.proxy.receivers.Package pack = this.getNextPackage();
			ClientInfo info = pack.getClientInfo();
			ReceiverMessage msg = pack.getMessage();
			ReceiverResult res = this.process(info, msg);
			if (res != null && res.getDestination() != null) {
				pack.setDestination(res.getDestination());
				pack.setMessage(res.getMessage());
				disp.addPackage(pack);
			}
		}
	}
	
	/**
	 * Metodo de processamento individual de pacotes
	 * @param info informaes do destinatario
	 * @param msg mensagem
	 * @return contem mensagem processada e proximo receiver destino
	 */
	protected abstract ReceiverResult process(ClientInfo info, ReceiverMessage msg);

}
